#include "Evalua.h"
#include "GUI.h"



extern void* hInstance;

inline HINSTANCE GetInstance() { return (HINSTANCE)hInstance; }



const char* GUI::GetClassName(void)
{
	sprintf(ClassName,"EVALUA%08x",(unsigned long)(LONG_PTR)this);

	return (const char*)ClassName;
}



GUI::GUI(AudioEffect* effect)
{
	WNDCLASS wc;

	plug=(Evalua*)effect;

	rect.left  =0;
	rect.top   =0;
	rect.right =GUI_WINDOW_WDT;
	rect.bottom=GUI_WINDOW_HGT;

	wc.style=CS_GLOBALCLASS;
	wc.lpfnWndProc=(WNDPROC)WndProc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=GetInstance();
	wc.hIcon=0;
	wc.hCursor=LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground=GetSysColorBrush(COLOR_BTNFACE);
	wc.lpszMenuName=0;
	wc.lpszClassName=GetClassName();

	RegisterClass(&wc);

	HlCopy = false;
	HlPaste = false;
}



GUI::~GUI()
{
	UnregisterClass(GetClassName(),GetInstance());
}
	
	
const char evalua_hint_text[] =
"DEC and HEX(with 0x prefix) integer numbers allowed\n"
"+ - * / % && | ^ << >> ( ) arithmetic operators allowed (ADD,SUB,MUL,DIV,MOD,AND,OR,XOR,SHL,SHR)\n"
"! ~ &&&& || == != < > <= >= logic and ? : ternary operators allowed as well\n"
"P=pitch, depends on the note pitch, 256 per period;   S=sine-like pitch\n"
"T=time since note start (65536 per second)\n"
"V=velocity 0..127;   M=modulation wheel 0..127;   R=random 0..32767";

bool GUI::open(void* ptr)
{
	HINSTANCE hinst;
	int x,y,w,h,style;

	AEffEditor::open(ptr);

	InitCommonControls();

	hinst=GetInstance();

	hWndForm = CreateWindow(GetClassName(), "Evalua", WS_CHILD | WS_VISIBLE, 0, 0, (rect.right - rect.left), (rect.bottom - rect.top), HWND(ptr), 0, hinst, this);

	if(!hWndForm) return false;

	SetWindowLongPtr(hWndForm, GWLP_USERDATA, (LONG_PTR)this);

	style = WS_VISIBLE | WS_CHILD | SS_OWNERDRAW;

	w = 80; 
	h = 25;
	x = GUI_WINDOW_WDT - 10 - w;
	y = 5;

	hWndButtonGain = CreateWindow("static", "", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	x -= w + 10;

	hWndButtonPorta = CreateWindow("static", "", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	x -= w + 10;

	hWndButtonPoly = CreateWindow("static", "", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	x -= w + 10;

	hWndButtonTimeScale = CreateWindow("static", "", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	style = WS_VISIBLE | WS_CHILD | SS_OWNERDRAW | SS_NOTIFY;

	x = 10;
	w = 40;

	hWndButtonCopy = CreateWindow("static", "Copy", style, x, y, w, h, hWndForm, (HMENU)IDC_BUTTON_COPY, hinst, NULL);

	x += w + 10;

	hWndButtonPaste = CreateWindow("static", "Paste", style, x, y, w, h, hWndForm, (HMENU)IDC_BUTTON_PASTE, hinst, NULL);

	x += w + 10; 
	
	hWndButtonLerp = CreateWindow("static", "Lerp", style, x, y, w, h, hWndForm, (HMENU)IDC_BUTTON_LERP, hinst, NULL);

	x += w + 10;

	hWndButtonClamp = CreateWindow("static", "Clamp", style, x, y, w, h, hWndForm, (HMENU)IDC_BUTTON_CLAMP, hinst, NULL);

	style = WS_CHILD | WS_VISIBLE | ES_LEFT | WS_BORDER | ES_MULTILINE | ES_UPPERCASE | ES_AUTOVSCROLL;

	x = 10;
	w = GUI_WINDOW_WDT - x * 2;
	y += h + 5;
	h = 25 * 5;

	hWndEditData = CreateWindow("edit", "0", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	SendMessage(hWndEditData, EM_SETLIMITTEXT, MAX_PROGRAM_LEN, 0);

	SetWindowLongPtr(hWndEditData, GWLP_USERDATA, 0xdeadf00b);	//Reaper hack to prevent hijacking Space button

	style = WS_CHILD | WS_VISIBLE;

	y += h + 5;
	w = GUI_WINDOW_WDT - x * 2;
	h = 20;

	hWndStaticError = CreateWindow("static", "ST_U", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	style = WS_VISIBLE | WS_CHILD | SS_OWNERDRAW;

	x = (GUI_WINDOW_WDT - GUI_VIEW_WDT) / 2;
	y += h;
	w = GUI_VIEW_WDT;
	h = GUI_VIEW_HGT;

	hWndStaticView = CreateWindow("static", "", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	style = WS_CHILD | WS_VISIBLE;

	y += h + 5;
	h = 25 * 5;
	w = GUI_VIEW_WDT - 60 - 10;

	hWndStaticHint = CreateWindow("static", "ST_U", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	SetWindowText(hWndStaticHint, evalua_hint_text);

	style = WS_VISIBLE | WS_CHILD | SS_OWNERDRAW;
	w = 80; 
	h = 25;
	x = GUI_WINDOW_WDT - w - 10;
	y = GUI_WINDOW_HGT - h - 10;

	hWndButtonDMap = CreateWindow("static", "", style, x, y, w, h, hWndForm, NULL, hinst, NULL);

	hFontError = CreateFont(16, 0, 0, 0, FW_BOLD, FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_TT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, TEXT("Tahoma"));
	hFontData = CreateFont(20, 0, 0, 0, FW_NORMAL, FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_TT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, TEXT("Tahoma"));
	hFontButton = CreateFont(14, 0, 0, 0, FW_DONTCARE, FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_TT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, TEXT("Tahoma"));
	hFontHint = CreateFont(12, 0, 0, 0, FW_DONTCARE, FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_TT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, TEXT("Tahoma"));

	SendMessage(hWndStaticError, WM_SETFONT, (WPARAM)hFontError, TRUE);
	SendMessage(hWndEditData, WM_SETFONT, (WPARAM)hFontData, TRUE);
	SendMessage(hWndStaticHint, WM_SETFONT, (WPARAM)hFontHint, TRUE);

	hViewDC = CreateCompatibleDC(GetDC(hWndForm));
	hViewBM = CreateCompatibleBitmap(GetDC(hWndForm), GUI_VIEW_WDT, GUI_VIEW_HGT);

	SelectObject(hViewDC, GetStockObject(DC_BRUSH));
	SelectObject(hViewDC, GetStockObject(DC_PEN));
	SelectObject(hViewDC, hViewBM);
	SelectObject(hViewDC, hFontButton);
	SetBkMode(hViewDC, TRANSPARENT);

	MouseActiveWnd = 0;

	SetUpdate(true);

	return true;
}



void GUI::close()
{
	DestroyWindow(hWndStaticError); 
	DestroyWindow(hWndStaticHint);
	DestroyWindow(hWndButtonPoly);
	DestroyWindow(hWndButtonPorta);
	DestroyWindow(hWndButtonTimeScale);
	DestroyWindow(hWndButtonGain);
	DestroyWindow(hWndButtonLerp);
	DestroyWindow(hWndButtonClamp);
	DestroyWindow(hWndButtonCopy);
	DestroyWindow(hWndButtonPaste);
	DestroyWindow(hWndButtonDMap);
	DestroyWindow(hWndEditData);
	DestroyWindow(hWndStaticView);
	DestroyWindow(hWndForm);

	DeleteObject(hFontError);
	DeleteObject(hFontData);
	DeleteObject(hFontButton);
	DeleteObject(hFontHint);

	DeleteDC(hViewDC);
	DeleteObject(hViewBM);

	AEffEditor::close();
}



bool GUI::getRect(ERect **ppErect)
{
   *ppErect=&rect;

   return true;
}



void GUI::idle(void)
{
	int map,pgm;
	char buf[32];
	const char* noteNames[12] = { "C-","C#","D-","D#","E-","F-","F#","G-","G#","A-","A#","B-" };

	pgm=plug->ProgramIndex;

	if(NeedUpdateEdit)
	{
		NeedUpdateEdit=false;

		SetWindowText(hWndEditData, plug->Program[pgm].Data);
	}

	if (NeedUpdate)
	{
		NeedUpdate = false;

		const char* err = plug->Compile();

		SetWindowText(hWndStaticError, err ? err : "OK");

		RenderView();

		if (plug->Polyphony > 1)
		{
			snprintf(buf, sizeof(buf), "Poly: %u", plug->Polyphony);
		}
		else
		{
			snprintf(buf, sizeof(buf), "Mono");
		}
		SetWindowText(hWndButtonPoly, buf);

		snprintf(buf, sizeof(buf), "Porta: %u", plug->PortaSpeed);
		SetWindowText(hWndButtonPorta, buf);

		snprintf(buf, sizeof(buf), "Time: %1.2fx", plug->GetTimeScaleX());
		SetWindowText(hWndButtonTimeScale, buf);

		snprintf(buf, sizeof(buf), "Gain: %u", plug->OutputGain);
		SetWindowText(hWndButtonGain, buf);

		map = plug->Program[pgm].DMap;

		if (map < 24)
		{
			snprintf(buf, sizeof(buf), "Ch10:  ---");
		}
		else
		{
			snprintf(buf, sizeof(buf), "Ch10:  %s%i", noteNames[map % 12], (map - 24) / 12 + 1);
		}

		SetWindowText(hWndButtonDMap, buf);

		InvalidateRect(hWndStaticError, NULL, true);
		InvalidateRect(hWndButtonPoly, NULL, true);
		InvalidateRect(hWndButtonPorta, NULL, true);
		InvalidateRect(hWndButtonTimeScale, NULL, true);
		InvalidateRect(hWndButtonGain, NULL, true);
		InvalidateRect(hWndButtonLerp, NULL, true);
		InvalidateRect(hWndButtonClamp, NULL, true);
		InvalidateRect(hWndButtonCopy, NULL, true);
		InvalidateRect(hWndButtonPaste, NULL, true);
		InvalidateRect(hWndButtonDMap, NULL, true);
		InvalidateRect(hWndStaticView, NULL, true);

		plug->updateDisplay();
	}
}



void GUI::SetUpdate(bool program)
{
	NeedUpdate=true;	
	NeedUpdateEdit=program;
}



void GUI::RenderView(void)
{
	char str[16];

	SetDCBrushColor(hViewDC, GetSysColor(COLOR_BTNFACE));
	SetDCPenColor(hViewDC, GetSysColor(COLOR_BTNFACE));

	RoundRect(hViewDC, 0, 0, GUI_VIEW_WDT, GUI_VIEW_HGT, 0, 0);

	SetDCBrushColor(hViewDC, GetSysColor(COLOR_WINDOWTEXT));
	SetDCPenColor(hViewDC, GetSysColor(COLOR_WINDOWTEXT));

	int sx = 0;
	int sy = 0;
	VstInt64 time_cnt = 0;
	VstInt64 pitch_cnt = 0;
	VstInt64 step = 4 + (1 << plug->ViewScale);

	int pgm = plug->ProgramIndex;

	plug->Program[pgm].ev->SetVar('V', 128);
	plug->Program[pgm].ev->SetVar('M', 0);

	for (int t = 0; t < GUI_VIEW_WDT; ++t)
	{
		plug->Program[pgm].ev->SetVar('T', time_cnt);
		plug->Program[pgm].ev->SetVar('P', pitch_cnt);
		plug->Program[pgm].ev->SetVar('R', rand());
		plug->Program[pgm].ev->SetVar('S', (pitch_cnt&~0xff) + plug->SineTable[pitch_cnt & 0xff]);
		plug->Program[pgm].ev->SetVar('C', (pitch_cnt&~0xff) + plug->SineTable[(pitch_cnt + 64) & 0xff]);

		int n = (int)(plug->Program[pgm].ev->Solve() / (256 / GUI_VIEW_HGT));

		if (plug->Clamp)
		{
			if (n < 0) n = 0;
			if (n > GUI_VIEW_WDT) n = GUI_VIEW_WDT;
		}
		else
		{
			n %= GUI_VIEW_WDT;
		}

		MoveToEx(hViewDC, sx, GUI_VIEW_HGT - n, NULL);
		LineTo(hViewDC, sx, GUI_VIEW_HGT);

		time_cnt += step * 8;
		pitch_cnt += step;
		++sx;
	}

	sprintf(str, "%ix", 8-plug->ViewScale);

	SetTextColor(hViewDC, GetSysColor(COLOR_BTNFACE));

	for (int i = 0; i < 3; ++i)
	{
		for (int j = 0; j < 3; ++j)
		{
			TextOut(hViewDC, 5 - 1 + i, GUI_VIEW_HGT - 20 - 1 + j, str, (int)strlen(str));
		}
	}
	SetTextColor(hViewDC, GetSysColor(COLOR_WINDOWTEXT));

	TextOut(hViewDC, 5, GUI_VIEW_HGT - 20, str, (int)strlen(str));
}



bool GUI::IsMouseOver(HWND hWnd)
{
	RECT rect;

	GetWindowRect(hWnd, &rect);

	return PtInRect(&rect, MousePoint) ? true : false;
}



void GUI::CheckParamLimits(void)
{
	int pgm = plug->ProgramIndex;

	if (plug->Polyphony < 1) plug->Polyphony = 1;
	if (plug->Polyphony > MAX_SYNTH_CHANNELS) plug->Polyphony = MAX_SYNTH_CHANNELS;

	if (plug->PortaSpeed < 1) plug->PortaSpeed = 1;
	if (plug->PortaSpeed > MAX_PORTA_SPEED) plug->PortaSpeed = MAX_PORTA_SPEED;

	if (plug->TimeScale < 1) plug->TimeScale = 1;
	if (plug->TimeScale > MAX_TIME_SCALE) plug->TimeScale = MAX_TIME_SCALE;

	if (plug->OutputGain < 1) plug->OutputGain = 1;
	if (plug->OutputGain > MAX_GAIN) plug->OutputGain = MAX_GAIN;

	if (plug->Program[pgm].DMap < 23) plug->Program[pgm].DMap = 23;
	if (plug->Program[pgm].DMap > 191) plug->Program[pgm].DMap = 191;
}



LRESULT WINAPI GUI::WndProc(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam)
{
	LPDRAWITEMSTRUCT lpDIS;
	char tempstr[32];
	HWND phWnd;
	HDC phDC;
	GUI* gui;
	bool compile;
	int i, w, dx, len, val, max, pgm, color;
	RECT rect,recta;

	phWnd=(HWND)lParam;
	phDC =(HDC)wParam;
	
	gui=(GUI*)GetWindowLongPtr(hWnd,GWLP_USERDATA);

	if (!gui || !gui->plug || !phWnd) return DefWindowProc(hWnd, message, wParam, lParam);

	pgm = gui->plug->ProgramIndex;

	switch(message)
	{
		case WM_COMMAND:
		{
			if (HIWORD(wParam) == EN_CHANGE)	//any edit has been changed
			{
				compile = false;

				if (phWnd == gui->hWndEditData) { GetWindowText(phWnd, gui->plug->Program[pgm].Data, MAX_PROGRAM_LEN); compile = true; }

				if(compile)
				{
					gui->plug->Program[pgm].NeedCompile = true;
					gui->SetUpdate(false);
				}
			}
			
			if (HIWORD(wParam) == STN_CLICKED)
			{
				switch (LOWORD(wParam))
				{
					case IDC_BUTTON_COPY:
					{
						DWORD Len = GetWindowTextLength(gui->hWndEditData) + 1;

						HGLOBAL hDst = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, (Len + 1) * sizeof(char));
						
						if (hDst)
						{
							LPSTR gDst = (LPSTR)GlobalLock(hDst);

							if (gDst)
							{
								if (OpenClipboard(gui->hWndForm))
								{
									EmptyClipboard();

									GetWindowText(gui->hWndEditData, gDst, Len);
									gDst[Len] = 0;
									GlobalUnlock(hDst);

									SetClipboardData(CF_TEXT, hDst);
									CloseClipboard();
								}
							}
						}

						gui->HlCopy = true;
						gui->SetUpdate(false);
					}
					break;

					case IDC_BUTTON_PASTE:
					{
						OpenClipboard(gui->hWndForm);

						HGLOBAL hClipBuf = GetClipboardData(CF_TEXT);

						if (hClipBuf)
						{
							LPSTR gSrc = (LPSTR)GlobalLock(hClipBuf);

							if (gSrc != NULL)
							{
								SetWindowText(gui->hWndEditData, gSrc);
								GetWindowText(gui->hWndEditData, gui->plug->Program[pgm].Data, MAX_PROGRAM_LEN);
								GlobalUnlock(gSrc);

								gui->plug->Program[pgm].NeedCompile = true;
							}
						}

						CloseClipboard();

						gui->HlPaste = true;
						gui->SetUpdate(false);
					}
					break;

					case IDC_BUTTON_LERP:
					{
						gui->plug->Lerp ^= true;
						gui->SetUpdate(false);
					}
					break;

					case IDC_BUTTON_CLAMP:
					{
						gui->plug->Clamp ^= true;
						gui->SetUpdate(false);
					}
					break;
				}
			}
		}
		return 0;

		case WM_LBUTTONDOWN:
		{
			gui->MousePoint.x = GET_X_LPARAM(lParam);
			gui->MousePoint.y = GET_Y_LPARAM(lParam);

			ClientToScreen(hWnd, &gui->MousePoint);

			gui->MouseOrigin = gui->MousePoint;
			gui->MouseActiveWnd = 0;

			if (gui->IsMouseOver(gui->hWndButtonPoly)) gui->MouseActiveWnd = gui->hWndButtonPoly;

			if (gui->IsMouseOver(gui->hWndButtonPorta)) gui->MouseActiveWnd = gui->hWndButtonPorta;

			if (gui->IsMouseOver(gui->hWndButtonTimeScale)) gui->MouseActiveWnd = gui->hWndButtonTimeScale;

			if (gui->IsMouseOver(gui->hWndButtonGain)) gui->MouseActiveWnd = gui->hWndButtonGain;

			if (gui->IsMouseOver(gui->hWndButtonDMap)) gui->MouseActiveWnd = gui->hWndButtonDMap;

			if (gui->IsMouseOver(gui->hWndStaticView))
			{
				gui->plug->ViewScale = (gui->plug->ViewScale + 1) & 7;
				gui->SetUpdate(false);
			}

			if(gui->MouseActiveWnd)
			{
				SetCapture(hWnd);
			}
		}
		return 0;

		case WM_LBUTTONUP:
		{
			ReleaseCapture();

			gui->MouseActiveWnd = 0;
		}
		return 0;

		case WM_RBUTTONDOWN:
		{
			gui->MousePoint.x = GET_X_LPARAM(lParam);
			gui->MousePoint.y = GET_Y_LPARAM(lParam);

			ClientToScreen(hWnd, &gui->MousePoint);

			if (gui->IsMouseOver(gui->hWndButtonTimeScale))
			{
				gui->plug->TimeScale = MAX_TIME_SCALE / 2;
				gui->SetUpdate(false);
			}

			if (gui->IsMouseOver(gui->hWndButtonPoly))
			{
				gui->plug->Polyphony = 8;
				gui->SetUpdate(false);
			}

			if (gui->IsMouseOver(gui->hWndButtonPorta))
			{
				gui->plug->PortaSpeed = MAX_PORTA_SPEED;
				gui->SetUpdate(false);
			}

			if (gui->IsMouseOver(gui->hWndButtonGain))
			{
				gui->plug->OutputGain = NORMAL_GAIN;
				gui->SetUpdate(false);
			}

			if (gui->IsMouseOver(gui->hWndButtonDMap))
			{
				gui->plug->Program[pgm].DMap = 0;
				gui->SetUpdate(false);
			}

			if (gui->IsMouseOver(gui->hWndStaticView))
			{
				gui->plug->ViewScale = (gui->plug->ViewScale - 1) & 7;
				gui->SetUpdate(false);
			}
		}
		return 0;

		case WM_MOUSEMOVE:
		{
			gui->MousePoint.x = GET_X_LPARAM(lParam);
			gui->MousePoint.y = GET_Y_LPARAM(lParam);

			ClientToScreen(hWnd, &gui->MousePoint);

			if (gui->MouseActiveWnd !=0)
			{
				dx = gui->MousePoint.x - gui->MouseOrigin.x;
				
				gui->MouseOrigin = gui->MousePoint;

				if (gui->MouseActiveWnd == gui->hWndButtonPoly) gui->plug->Polyphony += dx / 5;
				
				if (gui->MouseActiveWnd == gui->hWndButtonPorta) gui->plug->PortaSpeed += dx / 2;

				if (gui->MouseActiveWnd == gui->hWndButtonTimeScale) gui->plug->TimeScale += dx / 2;

				if (gui->MouseActiveWnd == gui->hWndButtonGain) gui->plug->OutputGain += dx / 2;

				if (gui->MouseActiveWnd == gui->hWndButtonDMap) gui->plug->Program[pgm].DMap += dx / 2;

				gui->CheckParamLimits();
				gui->SetUpdate(false);
			}
		}
		return 0;

		case WM_MOUSEWHEEL:
		{
			gui->MousePoint.x = GET_X_LPARAM(lParam);
			gui->MousePoint.y = GET_Y_LPARAM(lParam);

			dx = GET_WHEEL_DELTA_WPARAM(wParam);

			if (gui->IsMouseOver(gui->hWndButtonPoly)) gui->plug->Polyphony += dx / 120;
			if (gui->IsMouseOver(gui->hWndButtonPorta)) gui->plug->PortaSpeed += dx / 60;
			if (gui->IsMouseOver(gui->hWndButtonTimeScale)) gui->plug->TimeScale += dx / 120;
			if (gui->IsMouseOver(gui->hWndButtonGain)) gui->plug->OutputGain += dx / 60;
			if (gui->IsMouseOver(gui->hWndButtonDMap)) gui->plug->Program[pgm].DMap += dx / 120;

			gui->CheckParamLimits();
			gui->SetUpdate(false);
		}
		return 0;

		case WM_DRAWITEM:
		{
			lpDIS = (LPDRAWITEMSTRUCT)lParam;

			if (lpDIS->hwndItem == gui->hWndStaticView)
			{
				BitBlt(lpDIS->hDC, lpDIS->rcItem.left, lpDIS->rcItem.top, lpDIS->rcItem.right, lpDIS->rcItem.bottom, gui->hViewDC, 0, 0, SRCCOPY);
			}
			else
			{
				HDC memDC = CreateCompatibleDC(lpDIS->hDC);
				HBITMAP memBM = CreateCompatibleBitmap(lpDIS->hDC, lpDIS->rcItem.right - lpDIS->rcItem.left, lpDIS->rcItem.bottom - lpDIS->rcItem.top);

				SelectObject(memDC, memBM);

				SelectObject(memDC, GetStockObject(DC_BRUSH));
				SelectObject(memDC, GetStockObject(DC_PEN));
				SelectObject(memDC, gui->hFontButton);

				SetBkMode(memDC, OPAQUE);

				recta.left = 0;
				recta.top = 0;
				recta.right = lpDIS->rcItem.right - lpDIS->rcItem.left;
				recta.bottom = lpDIS->rcItem.bottom - lpDIS->rcItem.top;

				SetDCBrushColor(memDC, GetSysColor(COLOR_WINDOW));
				SetDCPenColor(memDC, GetSysColor(COLOR_WINDOW));

				RoundRect(memDC, recta.left, recta.top, recta.right, recta.bottom, 0, 0);

				SetDCBrushColor(memDC, GetSysColor(COLOR_BTNFACE));
				SetDCPenColor(memDC, GetSysColor(COLOR_WINDOWTEXT));

				RoundRect(memDC, recta.left, recta.top, recta.right, recta.bottom, 5, 5);

				rect.top = recta.top + 2;
				rect.left = recta.left + 2;
				rect.right = recta.right - 2;
				rect.bottom = recta.bottom - 2;

				w = rect.right - rect.left;

				max = w;
				val = 0;

				if (lpDIS->hwndItem == gui->hWndButtonGain)
				{
					max = MAX_GAIN;
					val = gui->plug->OutputGain;
				}

				if (lpDIS->hwndItem == gui->hWndButtonPoly)
				{
					max = MAX_SYNTH_CHANNELS;
					val = gui->plug->Polyphony;
				}

				if (lpDIS->hwndItem == gui->hWndButtonPorta)
				{
					max = MAX_PORTA_SPEED;
					val = gui->plug->PortaSpeed;
				}

				if (lpDIS->hwndItem == gui->hWndButtonTimeScale)
				{
					max = MAX_TIME_SCALE;
					val = gui->plug->TimeScale;
				}

				if (lpDIS->hwndItem == gui->hWndButtonLerp)
				{
					max = 1;
					val = gui->plug->Lerp ? 1 : 0;
				}

				if (lpDIS->hwndItem == gui->hWndButtonClamp)
				{
					max = 1;
					val = gui->plug->Clamp ? 1 : 0;
				}

				if (lpDIS->hwndItem == gui->hWndButtonCopy)
				{
					if (gui->HlCopy)
					{
						max = 1;
						val = 1;

						gui->HlCopy = false;
						gui->SetUpdate(false);
					}
				}

				if (lpDIS->hwndItem == gui->hWndButtonPaste)
				{
					if (gui->HlPaste)
					{
						max = 1;
						val = 1;

						gui->HlPaste = false;
						gui->SetUpdate(false);
					}
				}

				w = w * val / max;

				rect.right = rect.left + w;

				SetDCBrushColor(memDC, GetSysColor(COLOR_SCROLLBAR));
				SetDCPenColor(memDC, GetSysColor(COLOR_SCROLLBAR));

				RoundRect(memDC, rect.left, rect.top, rect.right, rect.bottom, 3, 3);

				len = GetWindowTextLength(lpDIS->hwndItem);

				GetWindowTextA(lpDIS->hwndItem, tempstr, len + 1);

				SetBkMode(memDC, TRANSPARENT);

				SetTextColor(memDC, GetSysColor(COLOR_BTNFACE));

				recta.top += 5;

				for (i = 0; i < 9; ++i)
				{
					rect.left = recta.left - 1 + (i % 3);
					rect.top = recta.top - 1 + ((i / 3) % 3);
					rect.right = recta.right;
					rect.bottom = recta.bottom;

					DrawTextA(memDC, tempstr, len, &rect, DT_CENTER);
				}

				SetTextColor(memDC, GetSysColor(COLOR_WINDOWTEXT));
				DrawTextA(memDC, tempstr, len, &recta, DT_CENTER);

				BitBlt(lpDIS->hDC, lpDIS->rcItem.left, lpDIS->rcItem.top, lpDIS->rcItem.right, lpDIS->rcItem.bottom, memDC, 0, 0, SRCCOPY);

				DeleteObject(memBM);
				DeleteDC(memDC);
			}
		}
		return TRUE;

		case WM_CTLCOLORSTATIC:
		{
			color = GetSysColor(COLOR_BTNFACE);

			SetBkColor(phDC, color);
		}
		return (LRESULT)GetSysColorBrush(COLOR_BTNFACE);

		case WM_ERASEBKGND:
		{

		}
		return TRUE;
	}

	return DefWindowProc(hWnd,message,wParam,lParam);
}